# -*- coding: utf-8 -*-
"""The Spiffy MultiPanel class and its methods.
"""

from collections import defaultdict
from itertools import product, combinations
import matplotlib
import matplotlib.pyplot as plt
import matplotlib.gridspec as gs
import numpy as np
import string
import math

from collections import namedtuple
from typing import Dict, Tuple, Union, Optional, Iterable
import warnings


class MultiPanel(object):

    """
    The central object of the `multipanel` module. Initiates a figure with multiple panels.
    """

    def __init__(
        self,
        shape: Optional[Tuple[int, int]] = (2, 2),
        grid: Union[Iterable[Tuple], Iterable[int]] = None,
        labels: Union[bool, Iterable[str], Dict[str, tuple], np.array] = False,
        **kwargs
    ) -> None:
        """
        The ``MultiPanel`` object is basically a wrapper of matplotlib's ``GridSpec``,
        but tries to simplify some aspects of multi-panel figure generation, such as Figure labels
        and the layout of panels. Depending on the input, the layout is initialized in one of three ways:

        **OPTION 1: Initialization based on the** ``labels`` **parameter**

            The ``labels`` parameter can be passed in as a dictionary, mapping custom figure labels (e.g. 'a', 'b', 'c')
            to locations in the grid that are defined by Tuples (e.g. {'A': (0, range(2,5)} will make a plot in the
            first row spanning columns 2-4 and give it the label A.

            Similarly, ``labels`` can be passed as a 2-dimensional np.array of strings. In this case, the strings in
            the cells of the array correspond to the label of the panels. Adjacent identical labels are considered
            one panel. For example, the array::
                ['A', 'A', 'D']
                ['B', 'C', 'D']
                ['E', 'E', 'E']

            will create 5 panels, each occupying the space that the respective label takes up in the array.

            This option is useful when you want to control both the arrangement of panels, and the order and
            format of their labels. If label is passed in as a dictionary or np.array, the ``grid`` and ``shape``
            parameters are ignored.

        **OPTION 2: Initialization based on the** ``grid`` **parameter:**

            If option 1 does not apply, the class will try to be initialized through the ``grid`` parameter.

            Example:
                Generate a two-row figure with 3 columns (panels) in the first row and 2 columns (panels)
                in the second row::
                    >>> fig = MultiPanel(grid=[3, 2])

            Example:
                Generate a 2x3 figure with 5 panels, where one panel spans
                both rows in the last column::
                    >>> fig = MultiPanel(grid=[(0, 0), (0, 1), (1, 0), (1, 1), (range(0, 2), 2)])

        **OPTION 3: initialization based on the** ``shape`` **parameter:**

            if neither ``labels`` or ``grid`` are supplied, the class will generate one panel in each cell of the grid
            matrix, as defined by the ``shape`` parameter.

            Example:
                Generate a 3x3 grid with
                9 plots of equal size::
                    >>> fig = MultiPanel(shape=(3, 3))

        Args:
            shape (Tuple): Determines the shape of the MultiPanel grid layout.

            grid (Iterable[Tuple], Iterable[int]): Determines the layout of subplots across the MultiPanel matrix.
                Defaults to one plot in each cell of the ``shape`` matrix. Can be one of:

                * Iterable of grid location tuples of form ``[rows, columns]``, in which rows and columns are
                  either int (for a single cell) or Iterable (for spanning multiple cells).
                * Iterable of ints with length ``shape[0]``, which defines the number of plots in each row.
                  Each plot then has the size ``1 x shape[0]/int``.
                  **Attention**: ``shape[0]`` must be divisable by every element in ``grid``.

            labels (bool, Iterable[str], dict, np.array): Assigns labels to subplots. Defaults to False.
                Can be one of:

                * Boolean. If True, labels are assigned to plots first across rows, then across columns.
                * Iterable of strings assigning labels to subplots, in the same order as defined by ``grid``.
                * A Dictionary mapping [str] keys to [Tuple] locations in the grid. This setting overrides the grid.
                * A np.array of the same shape as ``shape``, mapping string names to the locations in the grid.
                  Figures can span multiple cells in the grid. Also overrides the grid.


        Keyword Args:
            figsize (Tuple): Size of the figure. Will be passed into ``matplotlib.pyplot.figure``.

            label_case (str): 'uppercase' or 'lowercase'.
                This and following kwargs are passed to ``MultiPanel._draw_labels``.
            label_weight (str): Weight of the figure labels. defaults to 'bold'
            label_size (int): Font Size for figure labels. defaults to 14.
            label_location (Tuple): Tuple. Location of the figure labels relative to axis origin.
                    Defaults to (-0.1, 1.1)

            left (float): left margin.
                This and following kwargs are passed to ``matplotlib.gridspec.GridSpec``
            right (float): right margin
            bottom (float): bottom margin
            top (float): top margin
            wspace (float): horizontal spacing
            hspace (float): vertical spacing
            width_ratios (Iterable): width ratios of columns
            height_ratios (Iterable): height ratios of rows
        """

        self.npanels = 0
        self.shape = shape
        self._locations = []
        self._labels = []
        self.panels = []

        # parse kwargs
        figsize = kwargs.pop("figsize", plt.rcParams.get("figure.figsize"))

        self.fig = plt.figure(figsize=figsize)

        # OPTION 1: INITIALIZATION BASED ON ``labels``
        # # # # # # # # # # # #
        # When labels is given as a numpy array or dictionary,
        # the shape and grid parameters are ignored.

        # If labels is given as a numpy array, decode it into dictionary form.
        if isinstance(labels, np.ndarray):
            labels = _decode_label_array(labels)

        if isinstance(labels, dict):

            # If other parameters were not passed as their default
            if grid is not None or shape != (2, 2):
                warnings.warn(
                    "``labels`` was provided as a dictionary or array."
                    "The input to ``grid`` and ``shape`` will be ignored."
                )

            # Set crucial variables
            self._labels = list(labels.keys())
            self._locations = list(labels.values())
            self.shape = _find_max_tuple(self._locations)
            self.npanels = len(self._labels)
            draw_labels = True

        else:

            # OPTION 2: INITIALIZATION BASED ON ``grid``
            # # # # # # # # # # # #

            if grid is not None:

                # OPTION 2.1: grid is passed as an Iterable of ints
                if all(isinstance(i, int) for i in grid):
                    self.shape, grid, self.npanels = _get_subplot_raster(grid)

                # OPTION 2.2: grid is passed as an Iterable of Tuples
                elif all(isinstance(i, Tuple) for i in grid):
                    self.npanels = len(grid)
                    self.shape = _find_max_tuple(grid)

                else:
                    raise TypeError(
                        "Sorry, ``grid`` is not a valid input. "
                        "Refer to the documentation for supported input types."
                    )

            # OPTION 3: INITIALIZATION BASED ON ``shape``
            # # # # # # # # # # # #
            else:
                # Make a panel at each cell of the grid defined by shape
                try:
                    self.shape = shape
                    self.npanels = int(np.prod(shape))

                except ValueError:
                    raise TypeError(
                        "Sorry, ``shape`` is not a valid input. "
                        "Refer to the documentation for supported input types."
                    )

                grid = list()
                for row in range(self.shape[0]):
                    for col in range(self.shape[1]):
                        grid.append((row, col))

            self._locations = grid

            # Get labels based on provided vector or revert to default
            if isinstance(labels, bool):
                self._labels = _get_letters(case=kwargs.pop("label_case", "uppercase"))[
                    : self.npanels
                ]
                draw_labels = labels

            elif isinstance(labels, Iterable):
                assert (
                    len(labels) == self.npanels
                ), "Length of label vector does not match number of panels."
                self._labels = list(labels)
                draw_labels = True

            else:
                raise TypeError(
                    "Sorry, ``labels`` is not a valid input. "
                    "Refer to the documentation for supported input types."
                )

        # MAKE SUBPLOT LAYOUT
        # # # # # # # # # # # #

        # Raise a warning if there are overlapping panels
        overlaps = _panel_overlap(self._locations, self.shape)
        if len(overlaps) != 0:
            warnings.warn(
                "One or more panel coordinates overlap: {}! You probably do not "
                "want this, double check your input coordinates."
            )

        # Initialize GridSpec and consider Keyword Arguments

        self.gridspec = gs.GridSpec(
            nrows=self.shape[0],
            ncols=self.shape[1],
            figure=self.fig,
            left=kwargs.pop("left", None),
            bottom=kwargs.pop("bottom", None),
            right=kwargs.pop("right", None),
            top=kwargs.pop("top", None),
            wspace=kwargs.pop("wspace", None),
            hspace=kwargs.pop("hspace", None),
            width_ratios=kwargs.pop("width_ratios", None),
            height_ratios=kwargs.pop("height_ratios", None),
        )

        Panels = namedtuple("Panels", [i for i in self._labels])
        self.panels = Panels(
            *[
                self.fig.add_subplot(_get_grid_location(loc, self.gridspec))
                for loc in self._locations
            ]
        )

        # If labels should be drawn, draw them now.
        if draw_labels:
            self._draw_labels(
                label_location=kwargs.pop("label_location", (-0.2, 1.1)),
                size=kwargs.pop("label_size", 14),
                weight=kwargs.pop("label_weight", "bold"),
            )

    def _draw_labels(
        self,
        label_location: Tuple[float, float] = (-0.2, 1.1),
        size: int = 14,
        weight: str = "bold",
    ) -> None:

        for ix in range(self.npanels):

            # make separate axis for label
            loc = self._locations[ix]
            axis_loc = (int(np.min(loc[0])), int(np.min(loc[1])))
            ax = self.fig.add_subplot(
                _get_grid_location(axis_loc, self.gridspec), label=self._labels[ix]
            )
            ax.axis("off")

            ax.text(
                label_location[0],
                label_location[1],
                self._labels[ix],
                transform=ax.transAxes,
                size=size,
                weight=weight,
                usetex=False,
                family="sans-serif",
            )


def _get_letters(case: Optional[str] = "uppercase") -> str:
    """

    :param case: 'lowercase' or 'uppercase'. Defaults to 'lowercase'.
    :return: string of ordered alphabet
    """
    if case == "lowercase":
        return string.ascii_lowercase
    else:
        return string.ascii_uppercase


def _is_iter_of_iters(labels) -> bool:
    """
    Helper function to check for iterable of iterables
    """
    return isinstance(labels, Iterable) and all(isinstance(_, Iterable) for _ in labels)


def _decode_label_array(labels: Iterable[Iterable]) -> dict:
    """
    Helper function to transform a numpy array of subplot specifications into a dictionary
    mapping labels to locations. Generally accepts iterables of iterables, including
    numpy arrays, list of lists, and list of strings, where the latter assumes
    labels are individual characters.

    :param labels: grid of labels that maps cells in in the grid to a subplot label
    :return: The mapping in dictionary form
    """

    # make sure we've got a list of lists
    if not _is_iter_of_iters(labels):
        raise TypeError(
            "Sorry, ``labels`` must be a iterable of iterables, where "
            "each sub-iterable is the same length"
        )

    label_grid = [list(_) for _ in labels]

    # verify labels format
    if not all(len(_) == len(label_grid[0]) for _ in label_grid[1:]):
        raise TypeError(
            "Sorry, ``labels`` must be a iterable of iterables, where "
            "each sub-iterable is the same length"
        )

    # collect grid positions for each label
    label_pos = defaultdict(list)
    for i, row in enumerate(label_grid):
        for j, label in enumerate(row):
            label_pos[label].append((i, j))

    # ensure labels spanning grid points are linear contiguous
    label_dict = {}
    for label, positions in label_pos.items():

        rows = list(set([_[0] for _ in positions]))
        cols = list(set([_[1] for _ in positions]))

        row_range = range(min(rows), max(rows) + 1)
        col_range = range(min(cols), max(cols) + 1)

        # check that the label grid positions form a box
        expected_coords = list(product(rows, cols))
        if set(positions) != set(expected_coords):
            raise TypeError(
                "Sorry, label grid spec contains invalid layout; "
                "all identical label positions must be adjacent"
            )

        if len(rows) == 1:
            row_range = rows[0]
        if len(cols) == 1:
            col_range = cols[0]

        label_dict[label] = (row_range, col_range)

    return label_dict


def _get_grid_location(
    location: Tuple, gridspec: matplotlib.gridspec.GridSpec
) -> matplotlib.gridspec.SubplotSpec:
    """
    From A tuple of locations in a grid, return the SubplotSpec at the given coordinates.

    :param location: Tuple of locations. Can take one of these forms:
                    - (int, int)
                    - (Iterable, int)
                    - (Iterable, Iterable)
    :param gridspec: matplotlib GridSpec object
    :return: matplotlib SubplotSpec object
    """
    rows, cols = location

    # if both are integers
    if isinstance(rows, int) and isinstance(cols, int):
        return gridspec[rows, cols]

    elif isinstance(rows, Iterable) and isinstance(cols, int):
        return gridspec[rows[0] : rows[-1] + 1, cols]

    elif isinstance(rows, int) and isinstance(cols, Iterable):
        return gridspec[rows, cols[0] : cols[-1] + 1]

    elif isinstance(rows, Iterable) and isinstance(cols, Iterable):
        return gridspec[rows[0] : rows[-1] + 1, cols[0] : cols[-1] + 1]


def _get_subplot_raster(
    grid: Iterable[int],
) -> Tuple[Tuple[int, int], Iterable[Tuple], int]:
    """
    Defines a subplot raster from an iterable of integers that defines the number of plots in each row.

    :param grid: Iterable of integers, defining the number of plots in each row.
                Length must equal ``grid[0]``

    :return:    - A Tuple defining the shape of the raster
                - A vector of tuples defining the locations of each plot in the grid
                - The number of panels
    """

    npanels = int(sum(grid))
    locations = []

    # calculate shape based on length of grid and least common multiple of grid
    shape = (len(grid), _lcm_of_array(grid))

    for row in range(len(grid)):

        # Size of each plot in this row
        size = shape[1] / grid[row]

        # Make tuples of locations of each plot
        for panel in range(grid[row]):
            if size == 1:
                locations.append((row, panel))
            else:
                locations.append(
                    (row, range(int(panel * size), int(panel * size + size)))
                )

    return shape, locations, npanels


def _lcm_of_array(a: Iterable[int]) -> int:
    """
    helper function to calculate the lowest common multiple of an array.
    :param a: Iterable array of integers
    :return: integer
    """

    lcm = a[0]
    for i in range(1, len(a)):
        lcm = lcm * a[i] // math.gcd(lcm, a[i])
    return lcm


def _find_max_tuple(
    x: Iterable[Tuple[Union[Iterable, int], Union[Iterable, int]]]
) -> Tuple[int, int]:
    """
    Given a list of integer / range tuples, returns the maximum values along the first and second dimension
    :param x: List of Tuples
    :return: Tuple of maximum values of first and second dimension that occur in x
    """
    max1 = np.max([np.max(i[0]) for i in x])
    max2 = np.max([np.max(i[1]) for i in x])

    # Add plus one to output to transform to dimensionality (i.e. a max value of 0 indicates 1 dimension)
    return max1 + 1, max2 + 1


def _panel_overlap(locations, shape=None):
    """
    Check a list of (x,y) location coordinates, which may be ranges, to ensure
    none overlap

    :param locations: list of (x,y) tuple locations
    :param shape: the shape the locations should fit into (deprecated)
    """

    # expand all coordinates for each location
    coords = []
    for loc in locations:
        xlocs = loc[0] if isinstance(loc[0], range) else [loc[0]]
        ylocs = loc[1] if isinstance(loc[1], range) else [loc[1]]
        coords.append(list(product(xlocs, ylocs)))

    # examine all pairs of locations to make sure nothing overlaps
    overlap = False
    for loc1, loc2 in combinations(coords, 2):
        overlap = set(loc1).intersection(loc2)
        if overlap:
            break

    return overlap



from matplotlib.figure import Figure as mpl_Figure
from ._panel import Panel


class Figure(mpl_Figure):
    """
    Top-level container for any figure.
    """

    def add_subplot(self, *args, **kwargs):
        """
        Overrides the add_subplot function to add a
        spiffyplots `Panel` object instead of a matplotlib `Axes` object.
        """



import numpy as np
import matplotlib.pyplot as plt
from matplotlib.collections import LineCollection
from typing import Dict, Tuple, Union, Optional, Iterable


def multiline(x: Iterable, y: Iterable, c: Iterable, axis=None, **kwargs):
    """
    Plot multiple lines color-coded with a colormap

    Args:
        x: Iterable of X coordinates
        y: Iterable of Y coordinates
        c: Iterable of CMAP values
        axis: matplotlib axis object

    Notes:
        len(xs) == len(ys) == len(c) is the number of line segments
        len(xs[i]) == len(ys[i]) is the number of points for each line (indexed by i)

    Returns
    -------
    lc : instance of matplotlib LineCollection object.
    """

    # Find axis if not defined
    if axis is None:
        axis = plt.gca()

    # Make a LineCollection Object
    segments = [np.column_stack([x, y]) for x, y in zip(xs, ys)]
    lc = LineCollection(segments, **kwargs)

    # set coloring of line segments
    #    Note: I get an error if I pass c as a list here... not sure why.
    lc.set_array(np.asarray(c))

    # add lines to axes and rescale
    #    Note: adding a collection doesn't autoscalee xlim/ylim
    ax.add_collection(lc)
    ax.autoscale()
    return lc



"""
SpiffyPlots base module.
This module contains class definitions that often inherit matplotlib's classes to
add spiffy functionality while still allowing for the original matplotlib syntax.
"""

from ._figure import Figure
from ._panel import Panel



"""Top-level package for SpiffyPlots."""

__author__ = """Julian Rossbroich"""
__email__ = "julian.rossbroich@fmi.ch"
__version__ = "0.6.0"

from .multipanel import MultiPanel
from .lineplots import multiline



"""
Defines the Panel object class that inherits from matplotlib.Axes.
"""
from matplotlib.axes import Axes

# Spiffy functions to call from axis objects
from .. import multiline


class Panel(Axes):
    def multiline(self, x, y, c, **kwargs):
        """
        Plot multiple lines color-coded with a colormap.
        See spiffyplots.lineplots.multiline for full documentation
        """
        return multiline(x, y, c, axis=self, **kwargs)



"""Tests for `spiffyplots.multipanel` module."""

from itertools import product

import unittest
import pytest
import spiffyplots.multipanel as mp

import matplotlib
import numpy as np


class TestMutiPanel(unittest.TestCase):
    """Tests for `spiffyplots` package."""

    def setUp(self):
        """
        Set up test fixtures for spiffyplots.multipanel
        """

    def tearDown(self):
        """Tear down test fixtures, if any."""

    def test_init_001_default(self):
        """
        Test initialization of MultiPanel object.

        001 - Default initialization with no parameters
        (Should create a 2x2 grid with 4 equal panels)
        """
        fig = mp.MultiPanel()

        # assert 4 panels
        self.assertEqual(fig.panels.__len__(), 4)
        self.assertEqual(fig._labels, "ABCD")
        self.assertEqual(fig.shape, (2, 2))

    def test_init_002_grid_intlist(self):
        """
        Test initialization of MultiPanel object.

        002 - Initialization based on ``grid`` being a list of integers that define the number of panels in each row.
        """

        grid = (3, 4, 4, 1)  # 12 panels
        fig = mp.MultiPanel(grid=grid)

        self.assertEqual(fig.panels.__len__(), 12)
        self.assertEqual(fig._labels, "ABCDEFGHIJKL")
        self.assertEqual(fig.shape, (4, 12))

        grid2 = (2, 1)  # 3 panels with location tuple test
        fig2 = mp.MultiPanel(grid=grid2)

        self.assertEqual(fig2.panels.__len__(), 3)
        self.assertEqual(fig2._labels, "ABC")
        self.assertEqual(fig2.shape, (2, 2))
        self.assertEqual(fig2._locations, [(0, 0), (0, 1), (1, range(0, 2))])

    def test_init_003_grid_tuples(self):
        """
        Test initialization of MultiPanel object.

        003 - Initialization based on ``grid`` being a list of location tuples.
        """

        grid = [(0, 0), (0, 1), (range(1, 3), 0), (range(1, 3), 1)]
        fig = mp.MultiPanel(grid=grid)

        self.assertEqual(fig.panels.__len__(), 4)
        self.assertEqual(fig._labels, "ABCD")
        self.assertEqual(fig.shape, (3, 2))
        self.assertEqual(fig._locations, grid)

    def test_init_004_labels_dict(self):
        """
        Test initialization of MultiPanel object.

        004 - Initialization based on ``labels`` being a dictionary
        mapping panel labels to locations.
        """

        labels = {
            "A1": (0, 0),
            "A2": (0, 1),
            "B": (range(1, 3), 0),
            "C": (range(1, 3), 1),
        }

        fig = mp.MultiPanel(labels=labels)

        self.assertEqual(fig.panels.__len__(), 4)
        self.assertEqual(fig._labels, list(labels.keys()))
        self.assertEqual(fig.shape, (3, 2))
        self.assertEqual(fig._locations, list(labels.values()))

    def test_init_005_labels_list(self):
        """
        Test initialization of MultiPanel object.

        005 - Initialization based on ``labels`` being a list of custom labels.
        """

        labels = ["A1", "A2", "B1", "B2"]
        fig = mp.MultiPanel(labels=labels)
        self.assertEqual(fig.panels.__len__(), 4)
        self.assertEqual(fig._labels, labels)

    def test_init_006_labels_array(self):
        """
        Test initialization of MultiPanel object.

        004 - Initialization based on ``labels`` being a numpy array label grid
        """

        labels = np.array([["A1", "A2"], ["B1", "B2"]])
        fig = mp.MultiPanel(labels=labels)
        self.assertEqual(fig.panels.__len__(), 4)
        self.assertEqual(set(fig._locations), set(product([0, 1], [0, 1])))

        labels = np.array([["A", "A"], ["B", "B"]])
        fig = mp.MultiPanel(labels=labels)
        self.assertEqual(set(fig._locations), set([(0, range(0, 2)), (1, range(0, 2))]))

        labels = np.array([["A", "B", "B"], ["C", "C", "C"], ["C", "C", "C"]])
        fig = mp.MultiPanel(labels=labels)

    def test_kwargs(self):
        """
        Test if different keyword arguments work as expected
        """

        # Add keyword arguments for label generator
        kwargs = {
            "label_case": "lowercase",
            "label_size": 10,
            "label_weight": "normal",
            "label_location": (-0.2, 1),
            # Add keyword arguments for figure size
            "figsize": (8, 8),
            # Add keyword arguments for gridspec
            "left": 0.1,
            "bottom": 0.1,
            "right": 1,
            "top": 1,
            "wspace": 0.1,
            "hspace": 0.1,
            "width_ratios": (1, 2),
            "height_ratios": (1, 2),
        }

        fig = mp.MultiPanel(**kwargs)

        self.assertEqual(fig._labels, "abcd")

    def test_errors_invalid_inputs(self):
        """
        Test TypeErrors if invalid inputs are given
        """
        self.assertRaises(TypeError, mp.MultiPanel, grid=[1, 2, "string"])

        self.assertRaises(TypeError, mp.MultiPanel, shape=(1, (2, 4)))

        self.assertRaises(TypeError, mp.MultiPanel, labels=123)

        # Too few labels for the number of panels
        self.assertRaises(AssertionError, mp.MultiPanel, grid=(1, 3), labels=["ABC"])

    def test_warnings(self):
        """
        Test Warnings if some conditions are met
        """

        # If parameters given are ignored
        self.assertWarns(
            Warning, mp.MultiPanel, labels={"A1": (0, 0), "A2": (0, 1)}, shape=(3, 2)
        )

        # If panels overlap
        self.assertWarns(
            Warning, mp.MultiPanel, labels={"A1": (0, 0), "A2": (0, range(2))}
        )


class Test_get_letters(unittest.TestCase):
    def test_lowercase(self):
        """Test _get_letters."""
        out = mp._get_letters(case="lowercase")
        self.assertEqual(out[2], "c")
        self.assertEqual(out[-1], "z")

    def test_uppercase(self):
        """Test _get_letters."""
        out = mp._get_letters(case="uppercase")
        self.assertEqual(out[2], "C")
        self.assertEqual(out[-1], "Z")

    def test_default(self):
        """Test _get_letters."""
        out = mp._get_letters()
        self.assertEqual(out[2], "C")
        self.assertEqual(out[-1], "Z")


class Test_is_iter_of_iters(unittest.TestCase):
    def test_default(self):
        """Test _is_iter_of_iters"""
        self.assertTrue(mp._is_iter_of_iters([[1]]))
        self.assertTrue(mp._is_iter_of_iters([[1], [2]]))
        self.assertTrue(mp._is_iter_of_iters(["ABC", "DEF"]))
        self.assertTrue(mp._is_iter_of_iters([]))
        self.assertFalse(mp._is_iter_of_iters(1))


class Test_decode_label_array(unittest.TestCase):
    def test_simple_array(self):
        grid_dict = mp._decode_label_array([["A", "B", "C"], ["D", "D", "D"]])
        self.assertTrue(grid_dict["D"] == (1, range(0, 3)))

        grid_dict = mp._decode_label_array([["A", "C", "E"], ["B", "D", "E"]])
        self.assertTrue(grid_dict["E"] == (range(0, 2), 2))

        grid_dict = mp._decode_label_array([["A", "C", "C"], ["B", "C", "C"]])
        self.assertTrue(grid_dict["C"] == (range(0, 2), range(1, 3)))

        grid_dict = mp._decode_label_array(
            [["A", "B", "B"], ["C", "C", "C"], ["C", "C", "C"]]
        )
        self.assertTrue(grid_dict["C"] == (range(1, 3), range(0, 3)))

        # discontiguous labels
        self.assertRaises(TypeError, mp._decode_label_array, [["A", "B"], ["B", "A"]])
        self.assertRaises(
            TypeError, mp._decode_label_array, [["A", "B", "C"], ["C", "B", "A"]]
        )

        # different types of iterable inputs
        grid_dict = mp._decode_label_array(["ABC", "DDD"])
        self.assertTrue(grid_dict["D"] == (1, range(0, 3)))

        grid_dict = mp._decode_label_array(np.array([["A", "B", "C"], ["D", "D", "D"]]))
        self.assertTrue(grid_dict["D"] == (1, range(0, 3)))

        # should raise on invalid input
        self.assertRaises(TypeError, mp._decode_label_array, 1)

    def test_complex_array(self):
        NotImplemented


class Test_get_grid_location(unittest.TestCase):
    def setUp(self):
        """
        Setup example gridspec object
        """
        self.grid = matplotlib.gridspec.GridSpec(3, 3)

    def test_ints(self):
        tuple = (0, 1)
        out = mp._get_grid_location(tuple, self.grid)

        self.assertEqual(out, self.grid[0, 1])

    def test_int_rowrange(self):
        tuple = (range(0, 2), 1)
        out = mp._get_grid_location(tuple, self.grid)

        self.assertEqual(out, self.grid[0:2, 1])

    def test_int_colrange(self):
        tuple = (1, range(0, 3))
        out = mp._get_grid_location(tuple, self.grid)

        self.assertEqual(out, self.grid[1, 0:3])

    def test_ranges(self):
        tuple = (range(0, 2), range(0, 3))
        out = mp._get_grid_location(tuple, self.grid)

        self.assertEqual(out, self.grid[0:2, 0:3])

    def test_lists(self):
        tuple = ([0, 1, 2], [0, 1])
        out = mp._get_grid_location(tuple, self.grid)

        self.assertEqual(out, self.grid[0:3, 0:2])


class Test_get_subplot_raster(unittest.TestCase):
    def setUp(self):
        """
        setup example grid and corresponding panel locations
        """
        self.grid = [2, 3, 1]
        self.locations = [
            (0, range(0, 3)),
            (0, range(3, 6)),
            (1, range(0, 2)),
            (1, range(2, 4)),
            (1, range(4, 6)),
            (2, range(0, 6)),
        ]

    def test_raster(self):
        out_shape, out_loc, out_panels = mp._get_subplot_raster(self.grid)

        self.assertEqual(out_shape, (3, 6))
        self.assertEqual(out_loc, self.locations)
        self.assertEqual(out_panels, 6)


class Test_lcm_of_array(unittest.TestCase):
    def setUp(self):
        """
        Setup example arrays
        """
        self.numpy = np.array([1, 2, 3, 4, 5])
        self.list = [2, 3, 1]

    def test_numpy(self):
        out = mp._lcm_of_array(self.numpy)
        self.assertEqual(out, 60)

    def test_list(self):
        out = mp._lcm_of_array(self.list)
        self.assertEqual(out, 6)


class Test_find_max_tuple(unittest.TestCase):
    def setUp(self):
        self.int_tuples = [(4, 6), (7, 3)]
        self.range_tuples = [(range(3), range(5)), (range(3, 7), range(2, 5))]
        self.long_mixed = [
            (range(3), range(5)),
            (10, 1),
            (range(3, 7), range(2, 5)),
            (range(4), 3),
        ]

    def test_int_tuples(self):
        out = mp._find_max_tuple(self.int_tuples)
        self.assertEqual(out, (8, 7))

    def test_range_tuples(self):
        out = mp._find_max_tuple(self.range_tuples)
        self.assertEqual(out, (7, 5))

    def test_long_mixed(self):
        out = mp._find_max_tuple(self.long_mixed)
        self.assertEqual(out, (11, 5))

    def test_large_subplots(self):
        grid_dict = mp._decode_label_array(
            [["A", "B", "B"], ["C", "C", "C"], ["C", "C", "C"]]
        )
        self.assertEqual(mp._find_max_tuple(grid_dict.values()), (3, 3))


class Test_panel_overlap(unittest.TestCase):
    def test_default(self):
        self.assertFalse(mp._panel_overlap([]))
        self.assertFalse(mp._panel_overlap([(0, 0)]))
        self.assertFalse(mp._panel_overlap([(0, 0), (0, 1)]))
        self.assertFalse(mp._panel_overlap([(0, 0), (1, 0), (0, 1), (1, 1)]))
        self.assertTrue(mp._panel_overlap([(0, 0), (1, 0), (0, 1), (1, 0)]))
        self.assertFalse(mp._panel_overlap([(0, range(0, 10)), (1, range(0, 10))]))
        self.assertTrue(mp._panel_overlap([(0, range(0, 2)), (0, range(1, 2))]))
        self.assertFalse(
            mp._panel_overlap([(range(0, 2), range(0, 2)), (range(2, 4), range(2, 4))])
        )
        self.assertTrue(
            mp._panel_overlap([(range(0, 2), range(0, 2)), (range(1, 4), range(1, 4))])
        )

    def test_large_subplots(self):
        grid_dict = mp._decode_label_array(
            [["A", "B", "B"], ["C", "C", "C"], ["C", "C", "C"]]
        )
        self.assertFalse(mp._panel_overlap(grid_dict.values(), (3, 3)))



#!/usr/bin/env python

"""Setup script for SpiffyPlots.

As this package only distributes matplotlib style sheets so far,
the setup simply copies the *.mplstyle files into the appropriate directory.

This code is based on a StackOverflow answer:
https://stackoverflow.com/questions/31559225/how-to-ship-or-distribute-a-matplotlib-stylesheet

"""

import atexit
import glob
import os
import shutil
import matplotlib

from setuptools import setup, find_packages
from setuptools.command.install import install

# Get description from README
root = os.path.abspath(os.path.dirname(__file__))
with open(os.path.join(root, "README.md"), "r", encoding="utf-8") as f:
    long_description = f.read()

# Install requirements
requirements = ["matplotlib", "numpy"]

# Test suite requirements
test_requirements = ["coverage", "pytest", "pytest-cov"]

extras = {
    "test": test_requirements,
}


def install_styles():
    # Find all style files
    stylefiles = glob.glob("styles/**/*.mplstyle", recursive=True)

    # Find stylelib directory (where the *.mplstyle files go)
    mpl_stylelib_dir = os.path.join(matplotlib.get_configdir(), "stylelib")
    if not os.path.exists(mpl_stylelib_dir):
        os.makedirs(mpl_stylelib_dir)

    # Copy files over
    print("Installing styles into", mpl_stylelib_dir)
    for stylefile in stylefiles:
        print(os.path.basename(stylefile))
        shutil.copy(
            stylefile, os.path.join(mpl_stylelib_dir, os.path.basename(stylefile))
        )


class PostInstallMoveFile(install):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        atexit.register(install_styles)


setup(
    name="spiffyplots",
    url="https://github.com/JRBCH/spiffyplots",
    version="0.6.0",
    author="Julian Rossbroich",
    author_email="julian.rossbroich@fmi.ch",
    license="GPL-3",
    description=(
        "A collection of matplotlib style sheets and plotting tools for"
        " publication-ready figures"
    ),
    long_description=long_description,
    long_description_content_type="text/markdown",
    packages=find_packages(),
    include_package_data=True,
    test_suite="tests",
    install_requires=requirements,
    tests_require=test_requirements,
    extras_require=extras,
    cmdclass={"install": PostInstallMoveFile},
)



"""
Plotting examples for SpiffyPlots
"""
import numpy as np
import matplotlib.pyplot as plt
from spiffyplots import MultiPanel

# Data
# # # # # #

data = {
    "scatter1-x": np.random.randn(200),
    "scatter1-y": np.random.randn(200),
    "scatter2-x": np.random.rand(200),
    "scatter2-y": np.random.rand(200),
    "hist-gauss": np.random.randn(500),
    "hist-gamma": np.random.gamma(5, 8, 500),
    "heatmap": np.random.rand(100).reshape(10, 10),
    "timeseries1": np.array(
        [np.sin(np.arange(0, 25, 0.1)) + np.random.randn(250) for _ in range(20)]
    ),
    "timeseries1-true": np.sin(np.arange(0, 25, 0.1)),
    "timeseries2": np.array(
        [np.cos(np.arange(0, 25, 0.1)) + np.random.randn(250) for _ in range(20)]
    ),
    "timeseries2-true": np.cos(np.arange(0, 25, 0.1)),
}

# Functions
# # # # # # # #


def hist(axis, dat, c="black"):
    axis.hist(dat, color=c)
    axis.set_xlabel("value")
    axis.set_ylabel("frequency")


def scatter(axis, x, y, c="black"):
    axis.scatter(x, y, alpha=0.6, facecolor=c)
    axis.set_xlabel("variable 1")
    axis.set_ylabel("variable 2")


def timeseries(axis, y, true=False, c="black"):
    x = np.arange(0, y.shape[1])
    std = np.std(y, 0)
    axis.plot(y.mean(0), zorder=1, color=c)
    if True:
        axis.plot(true, zorder=2, color="black")
    axis.fill_between(
        x, y.mean(0) - std, y.mean(0) + std, zorder=0, facecolor=c, alpha=0.3
    )
    axis.set_xlabel("time (s)")
    axis.set_ylabel("value")


# PLOTTING
colors = plt.rcParams["axes.prop_cycle"].by_key()["color"]

fig = MultiPanel(
    grid=(4, 2),
    figsize=(9, 4),
    labels=True,
    label_location=(-0.2, 1.05),
    label_size=10,
    wspace=0.35,
    hspace=0.35,
)

hist(fig.panels[0], data["hist-gauss"], colors[0])
scatter(fig.panels[1], data["scatter1-x"], data["scatter1-y"], colors[0])
timeseries(fig.panels[4], data["timeseries1"], data["timeseries1-true"], colors[0])

hist(fig.panels[2], data["hist-gamma"], colors[1])
scatter(fig.panels[3], data["scatter2-x"], data["scatter2-y"], colors[1])
timeseries(fig.panels[5], data["timeseries2"], data["timeseries2-true"], colors[1])

fig.fig.savefig("examples/multipanel_mpl.png")


plt.style.use("spiffy")
colors = plt.rcParams["axes.prop_cycle"].by_key()["color"]

fig = MultiPanel(
    grid=(4, 2),
    figsize=(9, 4),
    labels=True,
    label_location=(-0.2, 1.05),
    label_size=10,
    wspace=0.35,
    hspace=0.35,
)

hist(fig.panels[0], data["hist-gauss"], colors[0])
scatter(fig.panels[1], data["scatter1-x"], data["scatter1-y"], colors[0])
timeseries(fig.panels[4], data["timeseries1"], data["timeseries1-true"], colors[0])

hist(fig.panels[2], data["hist-gamma"], colors[1])
scatter(fig.panels[3], data["scatter2-x"], data["scatter2-y"], colors[1])
timeseries(fig.panels[5], data["timeseries2"], data["timeseries2-true"], colors[1])

fig.fig.savefig("examples/multipanel_spiffy.png")
